unit League;

interface

uses SysUtils, Classes, Graphics,
     TornTeamBox1;

type
    TLeaguePlayer = class
    private
        FName,
        FButtonID,
        FHandicap,
        FComment: string;
    published
        property Name:     string read FName     write FName;
        property ButtonID: string read FButtonID write FButtonID;
        property Handicap: string read FHandicap write FHandicap;
        property Comment:  string read FComment  write FComment;
    end;

    TLeaguePlayerList = class(TList)  // of TLeaguePlayer
    private
        procedure Put(Index: Integer; Player: TLeaguePlayer);
        function  Get(Index: Integer): TLeaguePlayer;
    public
        function  Add(Player: TLeaguePlayer): TLeaguePlayer; overload;
        function  Add: TLeaguePlayer; overload;

        function  FindName(Name: string): Integer;
        function  FindButton(ButtonID: string): Integer;

        property Items[Index: Integer]: TLeaguePlayer read Get write Put; default;
    end;

    TLeagueTeam = class    // used to store data about each remembered league team
    private
        FTeamName: string;
        FTeamID:   Integer;
        FHandicap,
        FComment:  string;
        FPlayers: TLeaguePlayerList;
    public
        function  BuildTeamName: string;
    published
        Constructor Create;
        Destructor Destroy; override;

        property TeamName:  string  read FTeamName  write FTeamName;
        property TeamID:    Integer read FTeamID    write FTeamID;
        property Handicap:  string  read FHandicap  write FHandicap;
        property Comment:   string  read FComment   write FComment;
        property Players:   TLeaguePlayerList read FPlayers write FPlayers;
    end;

    TLeagueList = class(TList)  // of TLeagueTeam
    private
//        FHighTeamId: Integer;  // tracks what unique ID number should be allocated to the next team added to the League
    private
        procedure Put(Index: Integer; LeagueTeam: TLeagueTeam);
        function  Get(Index: Integer): TLeagueTeam;
    public
        function  Add(LeagueTeam: TLeagueTeam): TLeagueTeam; overload;
        function  Add: TLeagueTeam; overload;

        procedure Clear; Override;

        function  BuildTeamName(TeamBox: TTeamBox): string;
        function  RememberTeam(TeamBox: TTeamBox): Integer;
        procedure UpdateTeam(TeamBox: TTeamBox; LeagueTeam: TLeagueTeam);

        function  FindAllInLeague(TeamBox: TTeamBox): Integer;
        function  FindAnyInLeague(TeamBox: TTeamBox): Integer;
        function  FindTeamByName(TeamBox: TTeamBox): Integer;
        function  FindPlayer(buttonid: string): TLeaguePlayer;
        function  FindID(buttonid: string; MemberDBExists: Boolean = False): string;
        function  FindTeamID(teamid: Integer): TLeagueTeam;
        function  FindTeamByButton(buttonid: string): string;

        property Items[Index: Integer]: TLeagueTeam read Get write Put; default;
//        property HighTeamId: Integer read FHighTeamId write FHighTeamId;
    end;

implementation

uses TornMain;

{ TLeaguePlayerList }

procedure TLeaguePlayerList.Put(Index: Integer; Player: TLeaguePlayer);
begin
    inherited Put(Index, Player);
end;

function TLeaguePlayerList.Get(Index: Integer): TLeaguePlayer;
begin
    Result := inherited Get(Index);
end;

function TLeaguePlayerList.Add(Player: TLeaguePlayer): TLeaguePlayer;
begin
    inherited Add(Player);
    Result := Player;
end;

function TLeaguePlayerList.Add: TLeaguePlayer;
begin
    Result := Add(TLeaguePlayer.Create);
end;

function TLeaguePlayerList.FindName(Name: string): Integer;
begin
    Result := Count - 1;
    while Result > -1 do begin
        if Self[Result].Name = Name then Break;
        Dec(Result);
    end;
end;

function TLeaguePlayerList.FindButton(ButtonID: string): Integer;
begin
    Result := Count - 1;
    while Result > -1 do begin
        if Self[Result].ButtonID = ButtonID then Break;
        Dec(Result);
    end;
end;

{ TLeagueTeam }

constructor TLeagueTeam.Create;
begin
    Players := TLeaguePlayerList.Create;
end;

destructor TLeagueTeam.Destroy;
begin
    Players.Free;
    inherited;
end;

// Return a "team name" for a team in the League list.
function TLeagueTeam.BuildTeamName: string;
begin
    if TeamName <> '' then
        Result := TeamName
    else if Players.Count = 0 then
        Result := 'Team ' + IntToStr(TeamId)
    else if Players.Count = 2 then
        Result := Players[0].Name + ' and ' + Players[1].Name
    else
        Result := Players[0].Name + '''s team'
end;

{ TLeagueList }

procedure TLeagueList.Put(Index: Integer; LeagueTeam: TLeagueTeam);
begin
    inherited Put(Index, LeagueTeam);
end;

function TLeagueList.Get(Index: Integer): TLeagueTeam;
begin
    Result := inherited Get(Index);
end;

function TLeagueList.Add(LeagueTeam: TLeagueTeam): TLeagueTeam;
begin
    inherited Add(LeagueTeam);
    Result := LeagueTeam;
end;

function TLeagueList.Add: TLeagueTeam;
var i: Integer;
    highteamid: Integer;
begin
    highteamid := 0;
    for i := 0 to Count - 1 do
        if highteamid < Items[i].TeamID then
            highteamid := Items[i].TeamID;

    Result := Add(TLeagueTeam.Create);
    Result.TeamID := highteamid + 1;
end;

procedure TLeagueList.Clear;
var i: Integer;
begin
    for i := 0 to Count - 1 do
        if Assigned(Self[i]) then
            Self[i].Free;

//    FHighTeamId := 1;
    inherited;
end;

// Return a "team name" for this team, based on the number of its Team box.
// This is either the actual team name pulled from the Team box,
// or the actual team name pulled form the League info,
// or is a catenation of all the team's player names.
function TLeagueList.BuildTeamName(TeamBox: TTeamBox): string;
var i: Integer;
begin
    if (TeamBox.TeamName <> '') and (TeamBox.TeamName <> 'Alias') then begin
        Result := TeamBox.TeamName;  // return the team name from the Teambox
        Exit;
    end;

    i := FindAnyInLeague(TeamBox);
    if i > -1 then begin
        Result := Self[i].TeamName;  // return the team name from the team's League entry
        if Result <> '' then Exit;
    end;

    Result := '';
    for i := 0 to TeamBox.Items.Count - 1 do begin  // loop through each player on the team
        if i > 0 then begin
            if i = TeamBox.Items.Count - 1 then
                Result := Result + ' and '
            else
                Result := Result + ', ';
        end;

        with TeamBox.Items.Item[i] do
            if SubItems[tbMemberAlias] = '' then  // if this player has no alias
                Result := Result + Caption  // use his Pack ID
            else
                Result := Result + SubItems[tbMemberAlias];  // else use his alias
    end;  // for i
end;

// Remember all the players on this team in the League.
function TLeagueList.RememberTeam(TeamBox: TTeamBox): Integer;
var LeagueTeam: TLeagueTeam;
begin
    Result := Count;
    LeagueTeam := Add;
    with LeagueTeam do begin
        TeamName := '';
//        TeamID   := FHighTeamId;
//        Inc(FHighTeamId);
        Handicap := TeamBox.Handicap;
    end;
    UpdateTeam(TeamBox, LeagueTeam);
    TeamBox.TeamName := BuildTeamName(TeamBox);
    LeagueTeam.TeamName := TeamBox.TeamName;
end;

// Remember any players that aren't already on this team in the League.
// Called by RememberClick and UpdateClick.
procedure TLeagueList.UpdateTeam(TeamBox: TTeamBox; LeagueTeam: TLeagueTeam);
var i: Integer;
    playerindex: Integer;
    alias, buttonid: string;
    player: TLeaguePlayer;
begin
    for i := 0 to TeamBox.Items.Count - 1 do begin  // for each player in the Team box,
        alias    := TeamBox.Items[i].SubItems[tbMemberAlias];
        buttonid := TeamBox.Items[i].SubItems[tbMemberID];
        playerindex := LeagueTeam.Players.FindButton(buttonid);
        if //(Length(buttonid) = 7) and (buttonid[4] in ['+', '-', '%']) and
           (Copy(alias, 1, 6) <> 'Adjust') and (playerindex = -1) then begin
//            with LeagueTeam.Players.Add(TLeaguePlayer.Create) do begin  // add the new player
                player := LeagueTeam.Players.Add(TLeaguePlayer.Create);  // add the new player
                player.Name     := alias;
                player.ButtonID := buttonid;
                player.Handicap := '';
                player.Comment  := '';
            end
        else if (playerindex <> -1) and (LeagueTeam.Players[playerindex].Name <> alias) then
            LeagueTeam.Players[playerindex].Name := alias;  // update the existing player's alias
    end;  // for i

    LeagueTeam.Handicap := TeamBox.Handicap;
    TeamBox.TeamName := LeagueTeam.TeamName;
    TeamBox.HandicapColor := clWindow;
    TeamBox.Color := clWindow;
end;

// Find this team in the League list.
// Returns a match iff all players in the box are listed in the League entry.
function TLeagueList.FindAllInLeague(TeamBox: TTeamBox): Integer;
var i, j:       Integer;
    match:      Boolean;  // so far, all members found have been matched in this team
    foundmember:Boolean;  // so far, we have not found a member on this team
    alias, buttonid: string;
begin
    Result := -1;

    for i := 0 to Count - 1 do begin  // for each team in the League,
        match := True;
        foundmember := False;
        for j := 0 to TeamBox.Items.Count - 1 do begin  // for each player in the Team box,
            alias    := TeamBox.Items[j].SubItems[tbMemberAlias];  // grab the player's alias
            buttonid := TeamBox.Items[j].SubItems[tbMemberID];     // grab the player's button id

            foundmember := foundmember or ((buttonid <> '') and (Copy(buttonid, 1, 6) <> 'Adjust'));
            if (Length(buttonid) = 7) and (buttonid[4] in ['+', '-', '%']) then
                match := match and (Self[i].Players.FindName(alias) <> -1) and
                                   (Self[i].Players.FindButton(buttonid) <> -1);

            if not match then Break;  // as soon as match becomes false, it means (a) we must have found a member; (b) it ain't gonna become true again
        end;  // for j
        if match and foundmember then begin
            Result := i;
            Exit;
        end;
    end;  // while i
end;

// Find this team in the League list.
// Returns a match if any player in the box is listed in the League.
// Returns the best matching team; i.e. the team who has the most number of players listed in the Team box.
function TLeagueList.FindAnyInLeague(TeamBox: TTeamBox): Integer;
var i, j: Integer;
    s:  string;
    bestindex, confidence: Integer;
    bestconfidence: Double;
begin
    bestindex := -1;
    bestconfidence := 0;

    for i := 0 to Count - 1 do begin  // for each team in the League,
        confidence := 0;
        for j := 0 to TeamBox.Items.Count - 1 do begin  // for each player in the Team box,
            s := TeamBox.Items[j].SubItems[tbMemberAlias];  // grab the player's alias
            if (s <> '') and ((Self[i].Players.FindName(s) <> -1)) then  // if the player has an alias and it's in this team,
                Inc(confidence);
            s := TeamBox.Items[j].SubItems[tbMemberID];  // grab the player's button id
            if (s <> '') and ((Self[i].Players.FindButton(s) <> -1)) then  // if the player has a ButtonID and it's in this team,
                Inc(confidence);
        end;  // for j
        if (Self[i].Players.Count > 0) and (bestconfidence < confidence / Self[i].Players.Count) then begin
            bestconfidence := confidence / Self[i].Players.Count;
            bestindex  := i;
        end;
    end;  // while i

    Result := bestindex;
end;

// Find this team name in the League list.
function TLeagueList.FindTeamByName(TeamBox: TTeamBox): Integer;
var i: Integer;
begin
    for i := 0 to Count - 1 do  // for each team in the League,
        if (Self[i].TeamName = TeamBox.TeamName) then begin
            Result := i;
            Exit;
        end;
    Result := FindAnyInLeague(TeamBox);  // we didn't match a name, so match on (some) players
end;

// Look up a Button ID in the League list, and return the matching player alias.
function TLeagueList.FindPlayer(buttonid: string): TLeaguePlayer;
var i, j: Integer;
begin
    Result := nil;
    for i := 0 to Count - 1 do begin
        j := Self[i].Players.FindButton(buttonid);
        if j <> -1 then begin
            Result := Self[i].Players[j];
            Exit;
        end;
    end;
end;

function TLeagueList.FindID(buttonid: string; MemberDBExists: Boolean): string;
var player: TLeaguePlayer;
begin
    Result := '';
    player := FindPlayer(buttonid);
    if Assigned(player) then
        Result := player.Name;
end;

function  TLeagueList.FindTeamID(teamid: Integer): TLeagueTeam;
var i: Integer;
begin
    Result := nil;
    for i := 0 to Count - 1 do
        if Items[i].TeamID = teamid then begin
            Result := Items[i];
            Break;
        end;
end;

// Look up a Button ID in the League list, and return the name of the team(s) that player is listed on.
function TLeagueList.FindTeamByButton(buttonid: string): string;
var i, j: Integer;
begin
    Result := '';
    for i := 0 to Count - 1 do begin
        j := Self[i].Players.FindButton(buttonid);
        if j <> -1 then
            Result := Result + Self[i].BuildTeamName + ', ';
    end;
    Result := Copy(Result, 1, Length(Result) - 2);
    if Result = '' then Result := buttonid;
end;

end.
