unit TornDatePicker;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms;

type
  TBoxedDate = class
  public
    Date: TDateTime;
  end;

  TDateRangeEvent = procedure(StartDate, EndDate: TDateTime) of object;

  TTornDatePicker = class(TCustomControl)  // need to be based on TCustomControl rather than TGraphicControl, to eliminate flicker on repaint
  private
    FGames: TList;
    FGameDuration: TDateTime;  // in days; i.e. 0.01 is 14.4 minutes.
    FMinValue: TDateTime;  // lowest and highest date/times in the whole list
    FMaxValue: TDateTime;
    FMinDate: TDateTime;  // date range for the control -- larger than FMinValue..FMaxValue by 10%
    FMaxDate: TDateTime;
    FStartSelect: TDateTime;  // start and end points of the selection, if any
    FEndSelect: TDateTime;
    FOnSelect: TDateRangeEvent;
    FDoubleClicked: Boolean;  // True if the user just double-clicked the control  

    procedure SetGameDuration(Value: TDateTime);
    function XToDate: TDateTime;

    procedure DoDoubleClick(Sender: TObject);
    procedure DoMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure DoMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure DoMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
  protected
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure AddGame(Value: TDateTime);
    procedure ClearGames;
  published
    property GameDuration: TDateTime read FGameDuration write SetGameDuration;

    property OnSelect: TDateRangeEvent read FOnSelect write FOnSelect;
  end;

procedure Register;

implementation

uses
  Math;

procedure Register;
begin
  RegisterComponents('Samples', [TTornDatePicker]);
end;

constructor TTornDatePicker.Create(AOwner: TComponent);
begin
  inherited;

  FGameDuration := 15 / 1440;  // 15 minutes
  FGames := TList.Create;
  FMinValue := 99999;
  FMaxValue := 0;
  FMinDate := 99999;
  FMaxDate := 0;
  FStartSelect := 0;
  FEndSelect := 99999;
  FDoubleClicked := False;

  OnMouseDown := DoMouseDown;
  OnMouseUp := DoMouseUp;
  OnMouseMove := DoMouseMove;
  OnDblClick := DoDoubleClick;
end;

destructor TTornDatePicker.Destroy;
begin
  ClearGames;
  FGames.Free;

  inherited;
end;

procedure TTornDatePicker.AddGame(Value: TDateTime);
var newdate: TBoxedDate;
begin
  newdate := TBoxedDate.Create;
  newdate.Date := Value;
  FGames.Add(newdate);

  if FMinValue > Value then FMinValue := Value;
  if FMaxValue < Value then FMaxValue := Value;

  FMinDate := FMinValue - (FMaxValue - FMinValue) * 0.1;
  FMaxDate := FMaxValue + (FMaxValue - FMinValue) * 0.1;
end;

procedure TTornDatePicker.ClearGames;
var i: Integer;
begin
  for i := FGames.Count - 1 downto 0 do
    TObject(FGames[i]).Free;

  FGames.Clear;

  FMinDate := 99999;
  FMaxDate := 0;
end;

procedure TTornDatePicker.DoDoubleClick(Sender: TObject);
begin
  if FGames.Count >= 1 then begin
    FStartSelect := FMinDate;
    FEndSelect := FMaxDate + FGameDuration;
  end else begin
    FMinDate := Trunc(Now);
    FMaxDate := Trunc(Now) + 86399 / 86400;
  end;

  FDoubleClicked := True;

  if Assigned(FOnSelect) then
    FOnSelect(FStartSelect, FEndSelect);

  Invalidate;
end;

procedure TTornDatePicker.DoMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if not FDoubleClicked then
    FStartSelect := ScreenToClient(Mouse.CursorPos).X / Width * (FMaxDate + FGameDuration - FMinDate) + FMinDate;
end;

procedure TTornDatePicker.DoMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if FDoubleClicked then
    FDoubleClicked := False
  else begin
    FEndSelect := ScreenToClient(Mouse.CursorPos).X / Width * (FMaxDate + FGameDuration - FMinDate) + FMinDate;

    // Pass that out to our events.
    if Assigned(FOnSelect) then begin
      if FStartSelect < FEndSelect then
        FOnSelect(FStartSelect, FEndSelect)
      else
        FOnSelect(FEndSelect, FStartSelect);
    end;

    Invalidate;
  end;
end;

procedure TTornDatePicker.DoMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  ShowHint := False;
  Hint := DateTimeToStr(XToDate);
  ShowHint := True;
end;

procedure TTornDatePicker.Paint;

  function DateToX(Value: TDateTime): Integer;
  begin
    Result := Round((Value - FMinDate) / (FMaxDate + FGameDuration - FMinDate) * Width);
  end;

var
  i: Integer;
  thisdate: TDateTime;
  thisx: Integer;
  daywidth: Double;
  dayname: string;
begin
    Canvas.Brush.Color := clWhite;
    Canvas.FillRect(Rect(0, 0,  Width, Height));  // clear the canvas

    Canvas.Brush.Color := clBlue;
    Canvas.FillRect(Rect(DateToX(FStartSelect), 0,  DateToX(FEndSelect), Height));  // paint the selected area

    daywidth := Width / (FMaxDate + FGameDuration - FMinDate);

    Canvas.Pen.Color := clGray;
    Canvas.Font.Color := clGray;
    if FGames.Count > 0 then
      for i := Trunc(FMinDate) to Trunc(FMaxDate + FGameDuration) do
      begin
        Canvas.MoveTo(DateToX(i), 0);
        Canvas.LineTo(DateToX(i), Height);

        dayname := FormatDateTime('dddd', i);
        if Canvas.TextWidth(dayname) >= daywidth then begin
          dayname := FormatDateTime('ddd', i);
          if (Canvas.TextWidth(dayname) >= daywidth) and (Length(dayname) >= 1) then
            dayname := dayname[1];
        end;

        if (FStartSelect < i) and (i + 1 < FEndSelect) then
          Canvas.Brush.Color := clBlue
        else
          Canvas.Brush.Color := clWhite;

        Canvas.TextOut(Round(daywidth * (i - FMinDate) + (daywidth - Canvas.TextWidth(dayname)) / 2), (Height - Canvas.TextHeight(dayname)) div 2, dayname);
      end;

    for i := 0 to FGames.Count - 1 do begin
      thisdate := TBoxedDate(FGames[i]).Date;
      if (FStartSelect < thisdate) and (thisdate < FEndSelect) then
        Canvas.Brush.Color := $0200FF00
      else
        Canvas.Brush.Color := $0280E080;

      thisx := DateToX(TBoxedDate(FGames[i]).Date);
      Canvas.FillRect(Rect(thisx, 0, thisx + Max(DateToX(FMinDate + FGameDuration), 1), Height));
    end;
end;

procedure TTornDatePicker.SetGameDuration(Value: TDateTime);
begin
    FGameDuration := Value;
    Invalidate;
end;

function TTornDatePicker.XToDate: TDateTime;
begin
  Result := ScreenToClient(Mouse.CursorPos).X / Width * (FMaxDate + FGameDuration - FMinDate) + FMinDate;
end;

end.
