unit ZoomReport;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Contnrs, Math, Lists;

type
  TColumnType = ( ctString, ctInteger );

  TZColumn = class(TObject)
    Heading:   string;  // store the heading text for each column
    Heading2:  string;  // Optional. Appears *above* Heading.
    Alignment: string;  // 'left', 'right' or 'center'
    ColumnType: TColumnType;
  end;

  TZColumns = class(TList)
  public
    procedure Put(Index: Integer; ZColumn: TZColumn);
    function  Get(Index: Integer): TZColumn;
    function  Add(AHeading, AAlignment: string): TZColumn;
    function  Find(heading: string): Integer;

    property Items[Index: Integer]: TZColumn read Get write Put; default;
  end;

  TZCell = class(TObject)
  public
    Text:  string;    // text data
    Color: TColor;    // optional background colour
    BarCell: Integer; // optional pointer to cell whose value we are to show as a horizontal chart bar
    BarColor: TColor; // color of optional horizontal chart bar
    Scatter: TIntList;// list of values to be shown as a quartile plot
    Custom: TObject;  // can hold whatever data the caller wants.  It's the caller's responsibility to free it via an OnCellDestroy.

    constructor Create;
    destructor Destroy; override;

    procedure Assign(ASource: TZCell);
    function Clone: TZCell;
    function AsInteger: Integer;
  end;

  TZRow = class(TList)
  private
    FHasData: Boolean;
  public
    procedure Put(Index: Integer; ZCell: TZCell);
    function  Get(Index: Integer): TZCell;
    function  Add(AText: string; AColor: TColor; barcell: Integer): TZCell; overload;
    function  Add(AText: string; AColor: TColor): TZCell; overload;
    function  Add(AText: string): TZCell; overload;

    property Items[Index: Integer]: TZCell read Get write Put; default;
    property HasData: Boolean read FHasData write FHasData;  // Set by the user.  Used by PurgeBlankRows, below.
  end;

  TZRows = class(TList)
  public
    procedure Put(Index: Integer; ZRow: TZRow);
    function  Get(Index: Integer): TZRow;

    function AddRow: TZRow;
    procedure PurgeBlankRows;

    property Items[Index: Integer]: TZRow read Get write Put; default;
  end;

  TZoomReport = class;  // forward
  TCalcBar = procedure(report: TZoomReport; row, col: Integer; max: Real; var fill: Real);  // callback to custom-set bar cell filledness
  TPaintBar = procedure(report: TZoomReport; row, col: Integer; canvas: TCanvas; rect: TRect; Color, BarColor: TColor);  // callback to custom-paint bar background
  TOnCellDestroy = procedure(cell: TZCell);  // callback to free Custom data held in cells

  TColumnSet = set of Byte;

  TZoomReports = class;  // forward

  TNumberStyle = (nsPlain, nsComma, nsBrief);  // nsComma: use a thousands separator.  nsBrief: convert larger numbers to 2k, 3M, etc.

  TZoomReport = class(TObject)
  private
    FTitle:     string;       // title for the whole report
    FColumns:   TZColumns;
    FRows:      TZRows;
    FMaxBarByColumn: Boolean;  // if true, scale bar charts in each column separately
    FMaxBarByRow: Boolean;
    FMinBar:    Real;  // for scaling bar charts against
    FMinBars:   TDoubleList;  // used if FMaxBarByColumn is true, to store a max for each column
    FMaxBar:    Real;  // for scaling bar charts against
    FMaxBars:   TDoubleList;  // used if FMaxBarByColumn is true, to store a max for each column
    FDescription: string;
    FOwner:     TZoomReports;
    FNumberStyle: TNumberStyle;
    FMultiColumnOK: Boolean;  // If true, it is OK to try to render this in two "columns" a la Word, thus doubling the row height and halving the width of each column.

    FOnCalcBar: TCalcBar;
    FOnPaintBar: TPaintBar;
    FOnCellDestroy: TOnCellDestroy;

    procedure DoText(canvas: TCanvas; var left: Integer; top, width, height: Integer; text, align, widest: string);
    procedure SetMaxBarByColumn(const Value: Boolean);
    procedure SetMaxBarByRow(const Value: Boolean);
  public
    constructor Create; reintroduce; overload;
    constructor Create(AOwner: TZoomReports; ATitle, AHeadings, AAlignment: string); overload; virtual;
    destructor Destroy; override;

    function Darken(color: TColor): TColor;
    function GetBackColor(Color: TColor; row: Integer): TColor;  // Background colour of row. Different for odd/even rows.
    function GetBarColor(cell: TZCell; row: Integer): TColor;
    function RowCol(row: Integer; col: string): TZCell;

    function ColumnsWithData: TColumnSet;
    procedure PurgeBlankColumns;
    procedure DeleteColumn(i: Integer);

    function Show(canvas: TCanvas; left, top, right, bottom: Integer): Integer;
    procedure SaveAsHtml(var F: TextFile; bars: Boolean = False);
    procedure SaveAsCsv(var F: TextFile; separator: char = ',');

    property OnCalcBar: TCalcBar read FOnCalcBar write FOnCalcBar;
    property OnPaintBar: TPaintBar read FOnPaintBar write FOnPaintBar;
    property OnCellDestroy: TOnCellDestroy read FOnCellDestroy write FOnCellDestroy;
  published
    property Title:   string    read FTitle    write FTitle;
    property Columns: TZColumns read FColumns  write FColumns;
    property Rows:    TZRows    read FRows     write FRows;
    property MaxBarByColumn: Boolean read FMaxBarByColumn write SetMaxBarByColumn;
    property MaxBarByRow: Boolean read FMaxBarByRow write SetMaxBarByRow;
    property MaxBar:  Real      read FMaxBar   write FMaxBar;
    property Description: string read FDescription write FDescription;
    property Owner: TZoomReports read FOwner;
    property NumberStyle: TNumberStyle read FNumberStyle write FNumberStyle;
    property MultiColumnOK: Boolean read FMultiColumnOK write FMultiColumnOK;
  end;

  TZoomReports = class(TList)  // of TZoomReport
  private
    FTitleFontColor,
    FTitleBackColor,
    FTextColor,
    FBackgroundColor,
    FOddColor,
    FBarColor:   TColor;
    FHeightUsed: Integer;
  protected
    procedure Put(Index: Integer; ZoomReport: TZoomReport);
    function  Get(Index: Integer): TZoomReport;
  public
    constructor Create; overload;
    destructor Destroy; override;

    procedure Clear; override;
    function Add(ATitle, AHeadings, AAlignment: string): TZoomReport;
    property Items[Index: Integer]: TZoomReport read Get write Put; default;

    procedure Show(canvas: TCanvas; left, top, right, bottom: Integer);
    procedure SaveAsHtml(var F: TextFile; bars: Boolean = False);
    procedure SaveAsCsv(var F: TextFile; separator: char = ',');
  published
    property TitleFontColor:  TColor  read FTitleFontColor  write FTitleFontColor;
    property TitleBackColor:  TColor  read FTitleBackColor  write FTitleBackColor;
    property TextColor:       TColor  read FTextColor       write FTextColor;
    property BackgroundColor: TColor  read FBackgroundColor write FBackgroundColor;
    property OddColor:        TColor  read FOddColor        write FOddColor;
    property BarColor:        TColor  read FBarColor        write FBarColor;
    property HeightUsed:      Integer read FHeightUsed;
  end;

  TZoomReportss = class(TList)
  protected
    procedure Put(Index: Integer; ZoomReports: TZoomReports);
    function  Get(Index: Integer): TZoomReports;
  public
    destructor Destroy; override;

    procedure Clear; override;
    function Add(ZoomReports: TZoomReports): TZoomReports;
    property Items[Index: Integer]: TZoomReports read Get write Put; default;
  end;

implementation

//uses Dialogs;

{ TZColumns -------------------------------------------------------------------}

procedure TZColumns.Put(Index: Integer; ZColumn: TZColumn);
begin
  inherited Put(Index, @ZColumn);
end;

function TZColumns.Get(Index: Integer): TZColumn;
begin
  Result := inherited Get(Index);
end;

function TZColumns.Add(AHeading, AAlignment: string): TZColumn;
begin
  Result := TZColumn.Create;
  Result.Heading   := AHeading;
  Result.Alignment := AAlignment;
  inherited Add(Result);
end;

function  TZColumns.Find(heading: string): Integer;
begin
  Result := Count - 1;
  while (Result >= 0) and (Self[Result].Heading <> heading) do
    Dec(Result);
end;

{ TZCell ----------------------------------------------------------------------}

constructor TZCell.Create;
begin
  inherited;

  Color := clNone;
  BarCell := -1;
  Scatter := nil;
end;

destructor TZCell.Destroy;
begin
  Scatter.Free;

  inherited;
end;

procedure TZCell.Assign(ASource: TZCell);
begin
  Text := ASource.Text;
  Color := ASource.Color;
  BarCell := ASource.BarCell;
  BarColor := ASource.BarColor;
  Scatter := nil;  // I choose not to copy this: a shallow copy would be problematic if the owner of this cell frees it, and I can't be arsed doing a deep copy.
  Custom := ASource.Custom;  // This is an object, but as it's the caller's responsibility to free, we can just copy the pointer, not the object, with impunity.
end;

function TZCell.Clone: TZCell;
begin
  Result := TZCell.Create;
  Result.Assign(Self);
end;

function TZCell.AsInteger: Integer;
begin
  Result := StrToIntDef(Text, 0);
end;

{ TZRow -----------------------------------------------------------------------}

procedure TZRow.Put(Index: Integer; ZCell: TZCell);
begin
  inherited Put(Index, ZCell);
end;

function TZRow.Get(Index: Integer): TZCell;
begin
  Result := inherited Get(Index);
end;

function TZRow.Add(AText: string; AColor: TColor; barcell: Integer): TZCell;
begin
  Result := TZCell.Create;
  Result.Text    := AText;
  Result.Color   := AColor;
  Result.BarCell := barcell;
  Result.BarColor:= clNone;
  inherited Add(Result);
end;

function TZRow.Add(AText: string; AColor: TColor): TZCell;
begin
  Result := Add(AText, AColor, -1);
end;

function TZRow.Add(AText: string): TZCell;
begin
  Result := Add(AText, clNone, -1);
end;

{ TZRows ----------------------------------------------------------------------}

procedure TZRows.Put(Index: Integer; ZRow: TZRow);
begin
  inherited Put(Index, ZRow);
end;

function TZRows.Get(Index: Integer): TZRow;
begin
  Result := inherited Get(Index);
end;

function TZRows.AddRow: TZRow;
begin
  Result := TZRow.Create;
  Add(Result);
end;

procedure TZRows.PurgeBlankRows;
var i: Integer;
begin
  for i := Count - 1 downto 0 do
    if not Get(i).HasData then begin
      Get(i).Free;
      Delete(i);
    end;
end;

{ TZoomReport -----------------------------------------------------------------}

constructor TZoomReport.Create;
begin
  inherited;

  FColumns := TZColumns.Create;
  FRows := TZRows.Create;
  FMinBar := 0.0;
  FMinBars := TDoubleList.Create;
  FMaxBar := 0.0;
  FMaxBars := TDoubleList.Create;
  FNumberStyle := nsComma;
end;

constructor TZoomReport.Create(AOwner: TZoomReports; ATitle, AHeadings, AAlignment: string);
var i: Integer;
    headings, alignment: TStringList;
begin
  Create;

  FOwner   := AOwner;
  FTitle   := ATitle;

  headings            := TStringList.Create;
  headings.CommaText  := AHeadings;
  alignment           := TStringList.Create;
  alignment.CommaText := AAlignment;

  for i := 0 to Headings.Count - 1 do
    if i < alignment.Count then
      FColumns.Add(headings[i], alignment[i])
    else
      FColumns.Add(headings[i], 'right');

  headings.Free;
  alignment.Free;
end;

destructor TZoomReport.Destroy;
begin
  FColumns.Free;
  FRows.Free;     // this automagically frees all the Row's and cells.
  FMaxBars.Free;
  FMinBars.Free;

  inherited Destroy;
end;

function TZoomReport.Darken(color: TColor): TColor;
begin
  color  := ColorToRGB(color);
  Result := $02000000 + color and $FF  * 5 div 6 +
                    ((color and $FF00) * 5 div 6) and $FF00 +
                  ((color and $FF0000) * 5 div 6) and $FF0000;
end;

function TZoomReport.GetBackColor(Color: TColor; row: Integer): TColor;
begin
  if Color = clNone then
    if row mod 2 = 0 then Result := FOwner.BackgroundColor
                     else Result := FOwner.OddColor
  else
    Result := Color;
end;

function TZoomReport.GetBarColor(cell: TZCell; row: Integer): TColor;
begin
  if cell.BarColor <> clNone then
    Result := cell.BarColor
  else if (cell.Color = clNone) and (FOwner.BarColor <> clNone) then
    Result := FOwner.BarColor + GetBackColor(cell.Color, row) - $FFFFFF
  else
    Result := Darken(GetBackColor(cell.Color, row))
end;

// Return a set showing which columns contain data.  Any column whose index does not appear in this set is entirely blank.
function TZoomReport.ColumnsWithData: TColumnSet;
var j, k: Integer;
    maxcols: Integer;
    found: Boolean;
begin
  maxcols := 0;
  for j := 0 to Rows.Count - 1 do  // for each row
    maxcols := Max(maxcols, Rows[j].Count);

  Result := [];
  if Rows.Count > 0 then
    for k := 0 to maxcols - 1 do begin  // for each column
      found := False;
      for j := 0 to Rows.Count - 1 do  // for each row
        if (k < Rows[j].Count) and (Rows[j][k].Text <> '') then
        begin
          found := True;
          break;
        end;
      if found then
        Include(Result, k);
    end;
end;

procedure TZoomReport.PurgeBlankColumns;
var j, k: Integer;
    maxcols: Integer;
    datacols: TColumnSet;
begin
  datacols := ColumnsWithData;

  maxcols := 0;
  for j := 0 to Rows.Count - 1 do  // for each row
    maxcols := Max(maxcols, Rows[j].Count);

  for k := 0 to maxcols - 1 do  // for each column
    if not (k in datacols) then  // if it's up for deletion
       DeleteColumn(k);
end;

procedure TZoomReport.DeleteColumn(i: Integer);
var row, col: Integer;
begin
  if i < Columns.Count then
    Columns.Delete(i);

  for row := 0 to Rows.Count - 1 do begin  // for each row
    if i < Rows[row].Count then begin
      Rows[row][i].Free;
      Rows[row].Delete(i);
    end;

    for col := 0 to Rows[row].Count - 1 do
      if Rows[row][col].BarCell > i then
        Dec(Rows[row][col].BarCell);
  end;
end;

function TZoomReport.RowCol(row: Integer; col: string): TZCell;
var i: Integer;
begin
  i := Columns.Find(col);
  if (i > -1) and (row < Rows.Count) then
    Result := Rows[row][i]
  else
    Result := nil;
end;

// Convert a string to a float if it's valid, otherwise return a default value.
function StrToFloatDef(Val: string; Default: Real): Real;
begin
  if (Val = '') or (Val[1] in ['a'..'z', 'A'..'Z']) then
    Result := Default
  else
    try
      Result := StrToFloat(Val);
    except
      on EConvertError do Result := Default;
    end;
end;

function ToFloat(s: string): Real;
begin
  if Pos('%', s) > 1 then
    s := Copy(s, 1, Pos('%', s) - 1);
  Result := StrToFloatDef(s, 0.0);
end;

// Output a text field on the canvas; aligned left, centre or right; and with text sized to fit.
// Called by Show.
procedure TZoomReport.DoText(canvas: TCanvas; var left: Integer; top, width, height: Integer; text, align, widest: string);

  procedure ToBrief(var text: string);
  var val: Extended;
  begin
    if TextToFloat(PChar(text), val, fvExtended) then begin
      if (1000 <= val) and (val < 1000000) then
        text := IntToStr(Round(val / 1000.0)) + 'k'
      else if (1000000 <= val) then
        text := IntToStr(Round(val / 1000.0)) + 'M'
    end;
  end;

var s1, s2, s3, s4: string;
    fits1, fits2: Boolean;
begin
  case FNumberStyle of
    nsPlain: ;
    nsComma: if Abs(StrToFloatDef(text, 0.0)) > 999.9 then
                 text := FloatToStrF(StrToFloatDef(text, 0.0), ffNumber, 15, 0);  // reformat big numbers, adding thousands separators etc.
    nsBrief: ToBrief(text);
  end;

  // Find the best place to break the text across two lines.  Put the broken parts in s1 and s2.
  text := Trim(text);
  if Pos(' ', text) > 0 then begin
    s3 := Copy(text, 1, Pos(' ', text) - 1);
    s4 := Copy(text, Pos(' ', text) + 1, 999);
    s1 := s3;
    s2 := s4;
    while (canvas.TextWidth(s3) < canvas.TextWidth(s4)) and (Pos(' ', s4) > 0) do begin
      s1 := s3;
      s2 := s4;
      s3 := s3 + ' ' + Copy(s4, 1, Pos(' ', s4) - 1);
      s4 := Copy(s4, Pos(' ', s4) + 1, 999);
    end;
    if canvas.TextWidth(s2) > canvas.TextWidth(s3) then begin
      s1 := s3;
      s2 := s4;
    end;
  end else begin
    s1 := text;
    s2 := '';
  end;

  if (align <> 'right') or (canvas.TextWidth(widest) < canvas.TextWidth(text)) then
    widest := text;

  canvas.Font.Height := height;   // Font.Height is the height in pixels.  (Font.Size is size in _points_.)
  fits1 := True;
  while (canvas.Font.Height > 1) do begin
    fits1 := (canvas.TextWidth(widest + '  ') <= width) and (canvas.TextHeight(widest) <= height);  // does the text fits on one line at the current font height?
    fits2 := (canvas.TextWidth(s1 + '  ') <= width) and (canvas.TextWidth(s2 + '  ') <= width) and (canvas.TextHeight(s1) + canvas.TextHeight(s1) <= height);  // does the text fit on two lines in the current font height?
    if fits1 or fits2 then Break;
    canvas.Font.Height := canvas.Font.Height - 1;       // shrink the font by a pixel
  end;

  SetBkMode(canvas.Handle, TRANSPARENT);
  if fits1 then begin  // print it on a single line
    top := top + (height - canvas.TextHeight(text)) div 2;  // adjust the top so that the new small text is vertically aligned
    case align[1] of
      'l':  canvas.TextOut(left, top, text);                                          // left aligned
      'r':  canvas.TextOut(left +  width - canvas.TextWidth(text + ' '),  top, text); // right aligned
      else  canvas.TextOut(left + (width - canvas.TextWidth(text)) div 2, top, text); // centred
    end
  end else begin  // print it over two lines
    top := top + height div 2 - canvas.TextHeight(text);  // adjust the top so that the new small text is vertically aligned
    case align[1] of
      'l': begin  // left aligned
        canvas.TextOut(left, top, s1);
        canvas.TextOut(left, top + canvas.TextHeight(s1), s2);
      end;
      'r': begin  // right aligned
        canvas.TextOut(left +  width - canvas.TextWidth(s1 + ' '),  top, s1);
        canvas.TextOut(left +  width - canvas.TextWidth(s2 + ' '),  top + canvas.TextHeight(s1), s2);
      end;
      else begin  // centred
        canvas.TextOut(left + (width - canvas.TextWidth(s1)) div 2, top, s1);
        canvas.TextOut(left + (width - canvas.TextWidth(s2)) div 2, top + canvas.TextHeight(s1), s2);
      end;
    end;
  end;

  left := left + width;
  end;

// Show one report.  Return the bottom pixel of said report.
function TZoomReport.Show(canvas: TCanvas; left, top, right, bottom: Integer): Integer;
var currentleft, currenttop:  Integer;      // current position of the "cursor"
    colwidth, rowheight:      Integer;

  procedure QuartilePlot(cell: TZCell; maxcount: Integer; maxxbar: Real; backcolor, barcolor: TColor);
  var min, lower, median, upper, max, total: Real;
    i: Integer;
    c: Integer;  // cell.Scatter.Count
    markwidth: Integer;
    boxheight: Real;  // half the height of the box
    rowcentre: Real;  // vertical centre of the this row

    function Proport(x: Real): Integer;
    begin
      Result := currentleft + Round(colwidth * x / maxxbar);
    end;

  begin
    // paint cell background
    canvas.Brush.Color := backcolor;
    canvas.FillRect(Rect(currentleft, currenttop + 1, currentleft + colwidth, currenttop + rowheight));

    cell.Scatter.Sort;

    c := cell.Scatter.Count;

    if c > 0 then begin
      min    := cell.Scatter[0];
      lower  := cell.Scatter[c div 4];
      if c mod 2 = 0 then
        median := (cell.Scatter[c div 2 - 1] + cell.Scatter[c div 2]) / 2
      else
        median := cell.Scatter[c div 2];
      upper  := cell.Scatter[c * 3 div 4];
      max    := cell.Scatter[c - 1];

      total := 0;
      for i := 0 to c - 1 do
        total := total + cell.Scatter[i];

      markwidth := Math.Max(Round(colwidth / 200), canvas.TextWidth('l') div 2);
      rowcentre := currenttop + (rowheight + 1) / 2;
      boxheight := Sqrt(c / maxcount) * rowheight / 2;

      // paint bars
      canvas.Brush.Color := barcolor;
      canvas.FillRect(Rect(Proport(min),                Round(rowcentre - boxheight / 3),
                           Proport(lower),              Round(rowcentre + boxheight / 3)));  // bottom quartile
      canvas.FillRect(Rect(Proport(lower),              Round(rowcentre - boxheight),
                           Proport(median) - markwidth, Round(rowcentre + boxheight)));  // second quartile
      canvas.FillRect(Rect(Proport(median) + markwidth, Round(rowcentre - boxheight),
                           Proport(upper),              Round(rowcentre + boxheight)));  // third quartile
      canvas.FillRect(Rect(Proport(upper),              Round(rowcentre - boxheight / 3),
                           Proport(max),                Round(rowcentre + boxheight / 3)));  // top quartile

      // paint middle marks
//      canvas.Brush.Color := Darken(Darken(barcolor));
//      canvas.FillRect(Rect(Proport(median)- markwidth,  Round(rowcentre - boxheight),
//                           Proport(median) + markwidth, Round(rowcentre + boxheight)));  // median mark
//      canvas.FillRect(Rect(Proport(total / c) - markwidth, Round(rowcentre - boxheight),
//                           Proport(total / c) + markwidth, Round(rowcentre + boxheight)));  // mean mark

//      canvas.Brush.Color := clLime;
//      for i := 0 to c - 1 do
//        canvas.Ellipse(Round(Proport(cell.Scatter[i]) - boxheight),   Round(rowcentre - boxheight),
//                       Round(Proport(cell.Scatter[i]) + boxheight),   Round(rowcentre + boxheight));
    end;
  end;

var widths, widests:  TStringList;  // widths of each column; text of widest cell in each column
    maxcount:         Integer;  // largest number of scatter points in a single cell

  procedure ShowRow(j: Integer; scale: Real; left, right: Integer);
  var
    row:      TZRow;
    k:        Integer;
    barstart: Integer;  // starting and
    barend:   Integer;  // ending cells of a horizontal chart bar
    barcell:  Integer;  // cell in which the current bar's numeric value is stored
    fillfrom, fillto: Real; // 0.0 .. 1.0 portion of this chart bar to be filled
    minnbar, maxxbar: Real; // smallest and largest value used for a bar size, if horizontal bar charting is used
  begin
    canvas.Brush.Color := GetBackColor(clNone, j);
    canvas.FillRect(Rect(left, currenttop, right, currenttop + rowheight));
    row := Rows[j];

    // Paint the bar(s) in the background of this row.
    currentleft := left;
    barstart := 0;
    while barstart < Min(row.Count, Columns.Count) do begin
      if (row[barstart].BarCell <> -1) then begin  // there's a bar (or the start of a bar) to be painted in this cell
        // find the end of the bar
        barend := barstart;
        colwidth := 0;
        while (barend < widths.Count) and (barend < row.Count) and (row[barend].BarCell = row[barstart].BarCell) do begin
          colwidth := colwidth + Round(StrToFloatDef(widths[barend], 0.0) * scale);
          Inc(barend);
        end;

        // find the max value to scale the bar against
        if FMaxBarByColumn then begin
          minnbar := FMinBars[barstart];
          maxxbar := FMaxBars[barstart] * 1.01;
        end else if FMaxBarByRow then begin
          minnbar := FMinBars[j];
          maxxbar := FMaxBars[j] * 1.01;
        end else begin
          minnbar := FMinBar;
          maxxbar := FMaxBar;
        end;

        if minnbar > 0 then minnbar := 0;
        if maxxbar = 0 then maxxbar := 1;

        barcell := row[barstart].BarCell;

        fillfrom :=  -minnbar / (maxxbar - minnbar);
        fillto := (ToFloat(row[barCell].Text) - minnbar) / (maxxbar - minnbar);

        if fillfrom > fillto then begin
          fillfrom := fillto;
          fillto :=  -minnbar / (maxxbar - minnbar);
        end;

        if (fillfrom = fillto) and (colwidth > 0) then
          fillto := fillto + 1 / colwidth;

        if Assigned(FOnCalcBar) then
          FOnCalcBar(Self, j, barcell, maxxbar, fillto);  // let the caller customise the portion of the bar filled

        if Assigned(FOnPaintBar) then
          FOnPaintBar(Self, j, barcell, canvas, Rect(currentleft, currenttop, currentleft+colwidth, currenttop+rowheight),
                      GetBackColor(row[barcell].Color, j), GetBarColor(row[barcell], j))  // let the caller custom-paint the call background
        else if Assigned(row[barcell].Scatter) then
          QuartilePlot(row[barcell], maxcount, maxxbar, GetBackColor(row[barcell].Color, j), GetBarColor(row[barcell], j))
        else if (barstart = 0) or (fillto > 0) then begin
          canvas.Brush.Color := GetBarColor(row[barcell], j);
          canvas.FillRect(Rect(currentleft + Round(colwidth * fillfrom), currenttop+1,
                               currentleft + Round(colwidth * fillto), currenttop + rowheight));  // paint dark part of bar

          canvas.Brush.Color := GetBackColor(row[barcell].Color, j);
          canvas.FillRect(Rect(currentleft, currenttop+1,
                               currentleft + Round(colwidth * fillfrom), currenttop + rowheight));  // paint light part left of bar
          canvas.FillRect(Rect(currentleft + Round(colwidth * fillto), currenttop+1,
                               currentleft + colwidth, currenttop + rowheight));  // paint light part right of bar
        end else begin  // This is the fill < 0 case: the bar paints in the other direction.
          // paint right (dark) part of bar
          canvas.Brush.Color := GetBarColor(row[barcell], j);
          canvas.FillRect(Rect(currentleft + Round(colwidth * (1 + fillto)), currenttop+1, currentleft + colwidth, currenttop + rowheight));

          // paint left (light) part of bar
          canvas.Brush.Color := GetBackColor(row[barcell].Color, j);
          canvas.FillRect(Rect(currentleft, currenttop+1, currentleft + Round(colwidth * (1 + fillto)), currenttop + rowheight));
        end;

        barstart := barend;
        Inc(currentleft, colwidth);
      end else begin  // no bar to paint
        if (barstart < row.Count) and (row[barstart].Color <> clNone) then begin
          canvas.Brush.Color := row[barstart].Color;
          canvas.FillRect(Rect(currentleft, currenttop, currentleft + Round(StrToFloatDef(widths[barstart], 0.0) * scale), currenttop + rowheight));
        end;
        Inc(currentleft, Round(StrToFloatDef(widths[barstart], 0.0) * scale));
        Inc(barstart);
      end;
    end;  // while barstart

    // Paint the text in each of the cells in this row.
    currentleft := left;
    for k := 0 to Min(row.Count, Columns.Count) - 1 do
      DoText(canvas, currentleft, currenttop, Round(StrToFloatDef(widths[k], 0.0) * scale), rowheight,
             row[k].Text, Columns[k].Alignment, widests[k]);

    currenttop := currenttop + rowheight;
  end;

var j, k, m:          Integer;
    fontsize:         Integer;
    rowcount:         Integer;  // width of a column; number of rows in that column
    colwidest:        string;   // text of widest cell in a column
    totalwidth:       Real;
    thisbar:          Real;     // the current cell's bar size value, if horizontal bar charting is used
    barcell:          Integer;  // cell in which the current bar's numeric value is stored
    dummy:            Integer;
    descriptionrow:   Integer;  // 1 if there is a description; 0 otherwise.
    scale:            Real;
    multi:            Integer;  // Are we going to show this report in multiple sets of columns?
    hasheadings2:     Boolean;  // True if any column has Heading2 not blank.
    oldtop:           Integer;
    blockwidth, blockspace: Real;  // When presenting report in MultiColumn, these are how wide each column block is, and how wide the space between each column block.
begin  // TZoomReport.Show
  canvas.Font.Color  := FOwner.TitleFontColor;
  canvas.Font.Style  := [fsBold];
  canvas.Brush.Color := FOwner.TitleBackColor;
  canvas.Brush.Style := bsSolid;

  if Description <> '' then
    descriptionrow := 1
  else
    descriptionrow := 0;

  currenttop := top;
  maxcount := 0;

  // Find and remember the largest numeric value in each column.
  for j := 0 to Rows.Count - 1 do  // foreach row
    for k := 0 to Rows[j].Count - 1 do  // foreach col
      if (Rows[j][k].BarCell <> -1) then begin
        barcell := Rows[j][k].BarCell;

        if Assigned(Rows[j][barcell].Scatter) then begin
          thisbar := 0;
          for m := 0 to Rows[j][barcell].Scatter.Count - 1 do
            thisbar := Max(thisbar, Rows[j][barcell].Scatter[m]);  // get highest value fom the Scatter list

          if maxcount < Rows[j][barcell].Scatter.Count then
            maxcount := Rows[j][barcell].Scatter.Count;
        end else if Rows[j][barcell].BarCell > -1 then
          thisbar := ToFloat(Rows[j][Rows[j][barcell].BarCell].Text)  // get value of this cell
        else
          thisbar := 0;

        if FMinBar > thisbar then
          FMinBar := thisbar;  // store the smallest bar cell value in this report
        if FMaxBar < thisbar then
          FMaxBar := thisbar;  // store the largest bar cell value in this report

        if FMaxBarByColumn then begin
          while FMaxBars.Count <= k do begin
            FMinBars.Add(0.0);
            FMaxBars.Add(0.01);
          end;
          if FMinBars[k] > thisbar then
            FMinBars[k] := thisbar;  // store the smallest bar cell value in this column
          if FMaxBars[k] < thisbar then
            FMaxBars[k] := thisbar;  // store the largest bar cell value in this column
        end else if FMaxBarByRow then begin
          while FMaxBars.Count <= j do begin
            FMinBars.Add(0.0);
            FMaxBars.Add(0.01);
          end;
          if FMinBars[j] > thisbar then
            FMinBars[j] := thisbar;  // store the smallest bar cell value in this row
          if FMaxBars[j] < thisbar then
            FMaxBars[j] := thisbar;  // store the largest bar cell value in this row
        end;
      end;

  if FMaxBar = 0.0 then FMaxBar := 0.01;

  // Find relative widths for columns. For numeric columns, remember the biggest (which will also be the widest) value.
  widths  := TStringList.Create;
  widests := TStringList.Create;
  totalwidth := 0;
  canvas.Font.Size := 14;  // This is an arbitrary choice. Provided the font is more or less the same shape and kern across a range of point sizes (which it is), this is fine.
  for k := 0 to Columns.Count - 1 do begin
    if Columns[k].Alignment = 'right' then begin  // it's probably a numeric field
        colwidth := 0;
        rowcount := 1;
        for j := 0 to Rows.Count - 1 do  // loop through each cell in the column, and find the widest
          if (k < Rows[j].Count) and (StrToIntDef(Rows[j][k].Text, 0) > 999) and
             (colwidth < canvas.TextWidth(FloatToStrF(StrToFloatDef(Rows[j][k].Text, 0.0), ffNumber, 15, 0)))
          then begin
            colwidth  := canvas.TextWidth(FloatToStrF(StrToFloatDef(Rows[j][k].Text, 0.0), ffNumber, 15, 0) + '  ');
            colwidest := Rows[j][k].Text;
          end
          else if (k < Rows[j].Count) and (colwidth < canvas.TextWidth(Rows[j][k].Text)) then begin
            colwidth  := canvas.TextWidth(Rows[j][k].Text + '  ');
            colwidest := Rows[j][k].Text;
          end;
        widests.Add(colwidest);
    end else begin  // Alignment = left or center
      colwidth := 0;
      rowcount := 0;
      for j := 0 to Rows.Count - 1 do // loop through each cell in the column, and sum up the widths
        if k < Rows[j].Count then begin
          colwidth := colwidth + canvas.TextWidth(Rows[j][k].Text + '  ');
          Inc(rowcount);
        end;
      widests.Add('');
    end;  // if Alignment
    if rowcount = 0 then rowcount := 1;
    widths.Add(FloatToStr(colwidth / rowcount));
    totalwidth := totalwidth + colwidth / rowcount;
  end;  // for k

  if totalwidth = 0 then totalwidth := 1;
  scale := (right - left) / totalwidth;
  fontsize := Round(Abs(canvas.Font.Height) * 1.8 * scale);

  rowheight := (bottom - top) div (Rows.Count + 4 + descriptionrow);  // make an initial guess for row height
  if rowheight = 0 then rowheight := 1;

  if FMultiColumnOK and (rowheight < fontsize) then  // We are running out of width before we run out of height.
    multi := Trunc(SqRt(fontsize div rowheight))
  else
    multi := 1;
  rowheight := Trunc((bottom - top) / (Rows.Count / multi + 4 + descriptionrow));

  hasheadings2 := False;
  for k := 0 to Columns.Count - 1 do
    hasheadings2 := hasheadings2 or (Columns[k].Heading2 <> '');

  if hasheadings2 then
    canvas.FillRect(Rect(left, currenttop, right, currenttop + rowheight * 4 - 1))
  else
    canvas.FillRect(Rect(left, currenttop, right, currenttop + rowheight * 3 - 1));

  if Title <> '' then begin
    currenttop := currenttop + rowheight div 4;
    dummy := left;
    DoText(canvas, dummy, currenttop, right - left, rowheight * 3 div 2, Title, 'center', '');  // display the title, centred
    currenttop := currenttop + rowheight * 7 div 4;
  end;

  if multi > 1 then begin
    blockwidth := (right - left) / multi * 0.9;  // 90% of the width for blocks;
    blockspace := (right - left) / (multi - 1) * 0.1;  // 10% of the width for space between blocks.
    scale := scale / multi * 0.9;
  end else begin
    blockwidth := right - left;  // 100% of the width for the sole block.
    blockspace := 0;
  end;

  for m := 0 to multi - 1 do begin
    currentleft := Round(left + (blockwidth + blockspace) * m);
    for k := 0 to Columns.Count - 1 do  // Display Headings for this block.
      DoText(canvas, currentleft, currenttop, Round(StrToFloatDef(widths[k], 0.0) * scale), rowheight,
             Columns[k].Heading, Columns[k].Alignment, widests[k]);

    if hasheadings2 then begin
      currentleft := Round(left + (blockwidth + blockspace) * m);
      for k := 0 to Columns.Count - 1 do  // Display Heading2s.
        DoText(canvas, currentleft, currenttop + rowheight, Round(StrToFloatDef(widths[k], 0.0) * scale), rowheight,
               Columns[k].Heading2, Columns[k].Alignment, widests[k]);
    end;
  end;

  currenttop := currenttop + rowheight;
  if hasheadings2 then
    currenttop := currenttop + rowheight;

  canvas.Font.Color := FOwner.TextColor;
  canvas.Font.Style := [];
  oldtop := currenttop;
  for m := 0 to multi - 1 do begin  // For each block of rows,
    currenttop := oldtop;
    for j := Rows.Count * m div multi to Rows.Count * (m + 1) div multi - 1 do  // for each row that's in the block,
      ShowRow(j, scale, Round(left + (blockwidth + blockspace) * m),            // show it on canvas.
                        Round(left + (blockwidth + blockspace) * m + blockwidth));
  end;

  if Description <> '' then begin
    currentleft := left;
    DoText(canvas, currentleft, currenttop, right - left, rowheight, Description, 'left', Description);

    currenttop := currenttop + rowheight;
  end;
  Result := currenttop + rowheight;
  widths.Free;
  widests.Free;
end;

// Save one report as HTML to a file.
procedure TZoomReport.SaveAsHtml(var F: TextFile; bars: Boolean);
  function EncodeHTML(const s: string): string;
  begin
    Result := StringReplace(s, '&', '&amp;', [rfReplaceAll]);
    Result := StringReplace(Result, '<', '&lt;', [rfReplaceAll]);
    Result := StringReplace(Result, '>', '&gt;', [rfReplaceAll]);
  end;

  function BGRtoRGB(color: TColor): string;  // swap colours from BGR to RGB
  begin
    Result := Format('%.2x%.2x%.2x', [color and $FF, (color and $FF00) shr 8, (color and $FF0000) shr 16]);
  end;

    // True if each row has at most one bar.
//    function HasOneBar(report: TZoomReport) : Boolean;
//    var j, k: Integer;
//        barcell: Integer;
//    label endpoint;
//    begin
//        Result := True;
//        with report do
//            for j := 0 to Rows.Count - 1 do begin
//                barcell := Rows[j][0].BarCell;
//                for k := 1 to Rows[j].Count - 1 do begin
//                    if (barcell = -1) then
//                        barcell := Rows[j][k].BarCell;
//                    if (Rows[j][k].BarCell <> -1) and (Rows[j][k].BarCell <> barcell) then begin
//                        Result := False;
//                        goto endpoint;
//                    end;
//                end;
//            end;
//        endpoint:
//    end;

var j, k: Integer;
//    onebar: Boolean;
    cellcolor, barcolor: string;
    thisbar: Real;  // largest value used for a bar size, if horizontal bar charting is used
    datacols : TColumnSet;
    hasheadings2: Boolean;
begin
//  onebar := HasOneBar(Items[i]);
  datacols := ColumnsWithData;

  hasheadings2 := False;
  for k := 0 to Columns.Count - 1 do
    hasheadings2 := hasheadings2 or (Columns[k].Heading2 <> '');

  WriteLn(F, '');  // write the header of the table
  WriteLn(F, '<table align="center">');
  WriteLn(F, '  <thead>');
  if Title <> '' then begin
    WriteLn(F, '    <tr bgcolor="#' + BGRtoRGB(FOwner.TitleBackColor) + '">');
    WriteLn(F, '       <th colspan="' + IntToStr(Columns.Count) + '"><H2>' + Title + '</H2></th>');
    WriteLn(F, '    </tr>');
  end;

  if hasheadings2 then begin
    WriteLn(F, '    <tr bgcolor="#' + BGRtoRGB(FOwner.TitleBackColor) + '">');
    for k := 0 to Columns.Count - 1 do
      if k in datacols then
          WriteLn(F, '        <th align="' + Columns[k].Alignment + '">' + Columns[k].Heading2 + '</th>');
    WriteLn(F, '    </tr>');
  end;

  WriteLn(F, '    <tr bgcolor="#' + BGRtoRGB(FOwner.TitleBackColor) + '">');
  for k := 0 to Columns.Count - 1 do
    if k in datacols then
        WriteLn(F, '        <th align="' + Columns[k].Alignment + '">' + Columns[k].Heading + '</th>');
  WriteLn(F, '    </tr>');
  WriteLn(F, '  </thead>');
  WriteLn(F, '  <tbody>');

  for j := 0 to Rows.Count - 1 do begin // loop through each row in the report, and create a row on the canvas for it.
        // Write the <tr> tag that begins the row.
        (*if onebar then begin
            k := 0;
            while (k < Rows[j].Count) and (Rows[j][k].BarCell <> k) do
                Inc(k);
            if k < Rows[j].Count then begin
                if FMaxBarByColumn then
                    thisbar := ToFloat(FMaxBars[k])
                else
                    thisbar := FMaxBar;

                WriteLn(F, '    <tr class="bar' + BGRtoRGB(GetBackColor(Rows[j][k].Color, j)) +
                    '" style="background-position:' + IntToStr(Trunc((StrToFloatDef(Rows[j][k].Text, 0) * 100 / thisbar))) + '% 0">');
            end;
        end else*)
          WriteLn(F, '    <tr style="background-color: #' + BGRtoRGB(GetBackColor(clNone, j)) + '">');

      for k := 0 to Rows[j].Count - 1 do
        if k in datacols then begin
          // find colour for the cell
          if Rows[j][k].Color <> clNone then begin
            cellcolor := ' style="background-color: #' + BGRtoRGB(GetBackColor(Rows[j][k].Color, j)) + '"';
            barcolor :=  BGRtoRGB(Darken(GetBackColor(Rows[j][k].Color, j)));
          end else begin
            cellcolor := '';
            barcolor  := '';
          end;

          if bars and (*not onebar and*) (Rows[j][k].BarCell = k) then begin
            // find the max value to scale the bar against
            if FMaxBarByColumn then
              thisbar := FMaxBars[k]
            else if FMaxBarByRow then
              thisbar := FMaxBars[j]
            else
              thisbar := FMaxBar;

            WriteLn(F, '      <td><div class="barcontainer"' + cellcolor + '>');
            WriteLn(F, '        <div class="bar' + barcolor + '" style="width: ' + IntToStr(Trunc((ToFloat(Rows[j][k].Text) * 100 / thisbar))) + '%" />');
            WriteLn(F, '        <div align="' + Columns[k].Alignment + '" class="bartext">' +  EncodeHTML(Rows[j][k].Text) + '</div></div></td>');
//                        WriteLn(F, '      <td class="bar' + BGRtoRGB(Rows[j][k].Color) + '" align=' + Columns[k].Alignment + cellcolor +
//                            ' style="background-position:' + IntToStr(Trunc((StrToFloatDef(Rows[j][k].Text, 0) * 100 / thisbar))) + '% 0">' +  EncodeHTML(Rows[j][k].Text) + '</td>')
          end else
            WriteLn(F, '      <td align="' + Columns[k].Alignment + '"' + cellcolor + '>' + EncodeHTML(Rows[j][k].Text) + '</td>');
        end;  // if k; for k

      WriteLn(F, '    </tr>');
  end;  // for j to Rows.Count
  WriteLn(F, '</tbody></table>');
  if Description <> '' then
    WriteLn(F, '<p>' + Description + '</p>');
  WriteLn(F, '<br /><br />');
end;

// Take a list of items to show, and save 'em as CSV (comma separated values) to a file.
procedure TZoomReport.SaveAsCsv(var F: TextFile; separator: char);

  function QuoteString(const s: string): string;
  var foundquote: Boolean;
  begin
    Result := s;
    foundquote := Pos('"', s) > 0;

    if foundquote then Result := StringReplace(Result, '"', '""', [rfReplaceAll]);
    if foundquote or (Pos(separator, s) > 0) or (Pos(' ', s) > 0) or (Length(s) = 0) then
      Result := '"' + Result + '"';
  end;

var j, k: Integer;
    hasheadings2: Boolean;
begin
  hasheadings2 := False;
  for k := 0 to Columns.Count - 1 do
    hasheadings2 := hasheadings2 or (Columns[k].Heading2 <> '');

  for k := 0 to Columns.Count - 2 do
    Write(F, QuoteString(Columns[k].Heading) + separator);
  WriteLn(F, QuoteString(Columns[Columns.Count - 1].Heading));

  if hasheadings2 then begin
    for k := 0 to Columns.Count - 2 do
      Write(F, QuoteString(Columns[k].Heading2) + separator);
    WriteLn(F, QuoteString(Columns[Columns.Count - 1].Heading2));
  end;

  for j := 0 to Rows.Count - 1 do begin // loop through each row in the report, and create a row in the CSV for it
    for k := 0 to Rows[j].Count - 2 do
      Write(F, QuoteString(Rows[j][k].Text) + separator);
    WriteLn(F, QuoteString(Rows[j][Rows[j].Count - 1].Text));
  end;
end;

{ TZoomReports ----------------------------------------------------------------}

constructor TZoomReports.Create;
begin
  inherited Create;

  FTitleFontColor := clWhite;
  FTitleBackColor := clGray;
  FTextColor      := clBlack;
  FBackgroundColor:= clWhite;
  FOddColor       := $02FFFFEE;  // pale cyan
  FBarColor       := clNone;

  FHeightUsed := 0;
end;

destructor TZoomReports.Destroy;
begin
  Clear;

  inherited Destroy;
end;

procedure TZoomReports.Clear;
var i: Integer;
begin
  for i := 0 to Count - 1 do
    Items[i].Free;

  inherited Clear;
end;

function TZoomReports.Add(ATitle, AHeadings, AAlignment: string): TZoomReport;
begin
  Result := TZoomReport.Create(Self, ATitle, AHeadings, AAlignment);
  inherited Add(Result);
end;

procedure TZoomReports.Put(Index: Integer; ZoomReport: TZoomReport);
begin
  inherited Put(Index, ZoomReport);
end;

function TZoomReports.Get(Index: Integer): TZoomReport;
begin
  Result := inherited Get(Index);
end;

// Output reports to a canvas
procedure TZoomReports.Show(canvas: TCanvas; left, top, right, bottom: Integer);
var i:                Integer;
    currenttop:       Integer;      // current position of the "cursor"

begin
  currenttop := top;

  canvas.Font.Name := 'Tahoma';
  canvas.Brush.Color := FBackgroundColor;
  canvas.FillRect(Rect(left,top, right,bottom));  // clear the form

  for i := 0 to Count - 1 do
    currenttop := Items[i].Show(canvas, left, currenttop, right, currenttop + (bottom - currenttop) div (Count - i));

  FHeightUsed := currenttop;
end;

// Take a list of items to show, and save 'em as HTML to a file.
procedure TZoomReports.SaveAsHtml(var F: TextFile; bars: Boolean);

  function BGRtoRGB(color: TColor): string;  // swap colours from BGR to RGB
  begin
    Result := Format('%.2x%.2x%.2x', [color and $FF, (color and $FF00) shr 8, (color and $FF0000) shr 16]);
  end;

  // Return a list of colours of bar cells, over all reports.
  function BarCellColors: TStringList;
  var i, j, k: Integer;
  begin
    Result := TStringList.Create;
    Result.Sorted := True;
    Result.Duplicates := dupIgnore;
    for i := 0 to Count - 1 do  // for each report
      with Items[i] do
        for j := 0 to Rows.Count - 1 do
          for k := 1 to Rows[j].Count - 1 do
            if Rows[j][k].BarCell = k then
              Result.Add(BGRtoRGB(Darken(Rows[j][k].Color)));
  end;

var i: Integer;
//    onebar: Boolean;
    barcolors: TStringList;
begin
  Rewrite(F);
  try
    WriteLn(F, '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"');
    WriteLn(F, ' "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">');
    WriteLn(F, '');
    WriteLn(F, '<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">');
    WriteLn(F, '<head>');
    if Count > 0 then
      WriteLn(F, '  <title>' + Items[0].Title + '</title>');
    if bars then begin
      WriteLn(F, '  <style type="text/css">');
      WriteLn(F, '    .barcontainer { position:relative }');
      WriteLn(F, '    .bar { padding-bottom: 18px; background-color: #DFDFDF }');
      WriteLn(F, '    .bartext { position: absolute; top: 0px; left: 0px; text-align: right; width: 100% }');
      barcolors := BarCellColors;
      try
        for i := 0 to barcolors.Count - 1 do
          WriteLn(F, '    .bar' + barcolors[i] + ' { padding-bottom: 18px; background-color: #' + barcolors[i] + ' }');
      finally
        barcolors.Free;
      end;
      WriteLn(F, '  </style>');
    end;

    WriteLn(F, '</head>');
    WriteLn(F, '');
    WriteLn(F, '<body>');
    for i := 0 to Count - 1 do  // for each report
      Items[i].SaveAsHtml(F, bars);

    WriteLn(F, '</body>');
    WriteLn(F, '</html>');
  finally
    CloseFile(F);
  end;
end;

// Take a list of items to show, and save 'em as CSV (comma separated values) to a file.
procedure TZoomReports.SaveAsCsv(var F: TextFile; separator: char);
var i: Integer;
begin
  Rewrite(F);
  try
    for i := 0 to Count - 1 do
      Items[i].SaveAsCsv(F, separator);
  finally
    CloseFile(F);
  end;
end;

{ TZoomReportss }

destructor TZoomReportss.Destroy;
begin
  Clear;

  inherited;
end;

procedure TZoomReportss.Clear;
var i: Integer;
begin
  for i := 0 to Count - 1 do
    Items[i].Free;

  inherited Clear;
end;

function TZoomReportss.Add(ZoomReports: TZoomReports): TZoomReports;
begin
  Result := ZoomReports;
  inherited Add(Result);
end;

function TZoomReportss.Get(Index: Integer): TZoomReports;
begin
  Result := inherited Get(Index);
end;

procedure TZoomReportss.Put(Index: Integer; ZoomReports: TZoomReports);
begin
  inherited Put(Index, ZoomReports);
end;

procedure TZoomReport.SetMaxBarByColumn(const Value: Boolean);
begin
  FMaxBarByColumn := Value;

  if FMaxBarByColumn then FMaxBarByRow := False;
end;

procedure TZoomReport.SetMaxBarByRow(const Value: Boolean);
begin
  FMaxBarByRow := Value;

  if FMaxBarByRow then FMaxBarByColumn := False;
end;

end.
